unit GenVals;
{
	TGenVals component for Borland DELPHI.

	Written by Alexander Gogava,
        Tbilisi, Republic of Georgia. September 1999.

   E - mail   alex@argosoft.com

   http://www.argosoft.com/games/     (Home)
   http://www.argosoft.com/mathsim/   (Official)


         TGenVals (General Values) is Nonvisual
        Component - descendant from TComponenet.
	It can be use to solve General Eigen
        Values Problem (Determine frequencies
        and froms of natural vibrations) using
        Jacobi method by solution of a set of
        equations ([K]-Lambda[M])(X)=(0).
        (Min 2, Max 80 Equations).

 	 These component have been released to
        public domain, and may not be resold,
        without permission from the author.
	 You can use TGenVals in Your non
        commercial and commercial applications.
         If You use TGenVals, please send me
         a letter by the post mail at my job:

	Address:
	Alexander Gogava,
	Institute of Mechanics of Machines,
	Mindeli str. 10,
	Tbilisi-86,
	380086, Republic of Georgia

	 The author will not except any liability
        for bodily or mental harm that may have
        been taken in the use of these component.
	 All questions and comments are welcome.

	 List of Types, Public and Published
        properties of TGenVals happens to below.
	 Run demo program and see its source
        code, there demonstrated all main
        properties, event and function of the
        TGenVals.

Types:

  TMatr = array [1..80, 1..80] of Double;
  TVect = array [1..80] of Double;
  PMatr = ^TMatr;
  PVect = ^TVect;

Public Properties:

  property Mass: PMatr; - Matrix of Masses;
  property Stiff: PMatr; - Matrix of Sfiffnesses;

  property Forms: PMatr; - Matrix of Eigen Forms
  (Read Only);
  property Values: PVect; - Vector of Eigen
  Values (Read Only);

Published Properties:

  property Number: Word; - Number of equations.
  Default 2 (min 2 max 80). You can change number
  of equations only if property Active=False;
  property Active: Boolean; - It True, allocates
  the memory for matrixes. Set it True before use
  of component, set it False, to dispose memory
  (Default False);

Events:

 OnNotPositive : TNotifyEvent; - Occurs when
 masses or/and stiffnesses matrix is/are not
 positive definited;
 OnNotSymmetrical : TNotifyEvent; - Occurs when
 masses or/and stiffnesses matrix is/are not
 symmetrical;

Function:

 function CalcValues: Boolean; - Calculates
 values. Returns True, if calculation successful,
 Returns False, disposes memory and sets
 property Active to false, if calculation not
 successful.

Alexander Gogava

	 E - mail   alex@argosoft.com

	 http://www.argosoft.com/games/           (Home)
	 http://www.argosoft.com/mathsim/         (Official)
   }

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls;

type
  TMatr = array [1..80, 1..80] of Double;
  TVect = array [1..80] of Double;
  PMatr = ^TMatr;
  PVect = ^TVect;
  TGenVals = class(TComponent)
  private
  FNumber: Word;
  FActive: Boolean;
  FStiff, FMass, FForms: PMatr;
  FValues: PVect;
  FOnNotPositive: TNotifyEvent;
  FOnNotSymmetrical: TNotifyEvent;
  procedure SetActive(Value: Boolean);
  procedure SetNumber(Value: Word);
  public
  constructor Create(AOwner : TComponent); override;
  destructor Destroy; override;
  function CalcValues: Boolean;
  property Stiff: PMatr read FStiff write FStiff;
  property Mass: PMatr read FMass write FMass;
  property Forms: PMatr read FForms;
  property Values: PVect read FValues;
  published
  property Number: Word read FNumber write SetNumber default 2;
  property Active: Boolean read FActive write SetActive default False;
  property OnNotPositive : TNotifyEvent read FOnNotPositive write FOnNotPositive;
  property OnNotSymmetrical: TNotifyEvent read FOnNotSymmetrical write FOnNotSymmetrical;
  end;

procedure Register;

implementation

var
 D: PVect;

procedure Register;
begin
  RegisterComponents('Samples', [TGenVals]);
end;

destructor TGenVals.Destroy;
begin
  if D <> nil then begin Dispose (D); D:=nil; end;
  if FStiff<>nil then begin Dispose(FStiff); FStiff:=nil; end;
  if FMass<>nil then begin Dispose(FMass); FMass:=nil; end;
  if FForms<>nil then begin Dispose(FForms); FForms:=nil; end;
  if FValues<>nil then begin Dispose(FValues); FValues:=nil; end;
inherited Destroy;
end;

constructor TGenVals.Create(AOwner : TComponent);
begin
  D:=nil; FValues:=nil; FForms:=nil; FMass:=nil; FStiff:=nil;
  FNumber:=2; FActive:=False;
  inherited Create(AOwner);
end;

procedure TGenVals.SetNumber(Value: Word);
begin
  if FNumber <> Value then
    begin
    if FActive=True then
    begin
    MessageDlg ('You can change Number of equations only if property Active=False', mtError, [mbOK], 0);
    exit;
    end;
    if (Not (Value in [2..80]))then
    begin
    MessageDlg ('Number of equations must be from 2 to 80', mtError, [mbOK], 0);
    if (Not (csDesigning in ComponentState)) then FNumber:=Value;
    exit;
    end;
    FNumber:=Value;
    end;
end;

procedure TGenVals.SetActive(Value: Boolean);
var
i,j: Integer;
begin
if csDesigning in ComponentState then
begin
if FActive<>Value then FActive:=Value;
exit;
end;
  if Value = True then
  begin
    if (Not (FNumber in [2..80]))then
    begin
    MessageDlg ('Number of equations must be from 2 to 80', mtError, [mbOK], 0);
    exit;
    end;
  try
  if FStiff=nil then New(FStiff);
  if FMass=nil then New(FMass);
  if FForms=nil then New(FForms);
  if FValues=nil then New(FValues);
  except
  if FStiff<>nil then begin Dispose(FStiff); FStiff:=nil; end;
  if FMass<>nil then begin Dispose(FMass); FMass:=nil; end;
  if FForms<>nil then begin Dispose(FForms); FForms:=nil; end;
  if FValues<>nil then begin Dispose(FValues); FValues:=nil; end;
  exit;
  end;
  for i:=1 to FNumber do
  begin
  FValues^[i]:=0.0;
  for j:=1 to FNumber do
  begin
  FStiff^[i,j]:=0.0;
  FMass^[i,j]:=0.0;
  FForms^[i,j]:=0.0;
  end;
  end;
  end
  else
  begin
  if FStiff<>nil then begin Dispose(FStiff); FStiff:=nil; end;
  if FMass<>nil then begin Dispose(FMass); FMass:=nil; end;
  if FForms<>nil then begin Dispose(FForms); FForms:=nil; end;
  if FValues<>nil then begin Dispose(FValues); FValues:=nil; end;
  end;
FActive:=Value;
end;

function TGenVals.CalcValues: Boolean;
label Label1, Label2, Label3, Label4, Label5;
var
i, j, L, M, k, jj, j1, j2, k1, k2: Integer;
A, B, YY, XX, AB, C, X, Q, D1, D2, Z,
E, F, CA, CG, AJ, BJ, AK, BK, T, G, XA, XB, ZZ, TMP: Double;
begin
if D=nil then New(D); Result:=False;
for i := 1 to FNumber do
  begin
  if (FStiff^[i,i] <= 0) or (FMass^[i,i] <= 0) then
    begin
    if assigned (FOnNotPositive) then FOnNotPositive(Self);
    if D <> nil then begin Dispose (D); D:=nil; end;
    FActive:=False; Exit;
    end;
  D^[i] := FStiff^[i,i] / FMass^[i,i]; FValues^[i] := D^[i];
  end;
for i := 1 to FNumber do
for j := i to FNumber do
  if ((FStiff^[i,j] <> FStiff^[j,i]) or (FMass^[i,j] <> FMass^[j,i]))  then
    begin
    if assigned (FOnNotSymmetrical) then FOnNotSymmetrical(Self);
    if D <> nil then begin Dispose (D); D:=nil; end;
    FActive:=False; Exit;
    end;
for i := 1 to FNumber do
begin
for j := 1 to FNumber do
FForms^[i,j] := 0.0;
FForms^[i,i] := 1.0;
end;
if FNumber = 1 then
begin
if D <> nil then
begin Dispose (D); D:=nil; end;
FActive:=False;
end;;
L := 0;
M := FNumber - 1;
Label1:
Inc (L);
X:=Sqr(Exp(L*Ln(0.01)));
for j := 1 to M do
  begin
  jj := j + 1;
  for k := jj to FNumber do
  begin
  A := (FStiff^[j,k] * FStiff^[j,k]) / (FStiff^[j,j] * FStiff^[k,k]);
  B := (FMass^[j,k] * FMass^[j,k]) / (FMass^[j,j] * FMass^[k,k]);
  if (A < X) and (B < X) then goto Label2;
    XX := FStiff^[k,k] * FMass^[j,k] - FMass^[k,k] * FStiff^[j,k];
    YY := FStiff^[j,j] * FMass^[j,k] - FMass^[j,j] * FStiff^[j,k];
    AB := FStiff^[j,j] * FMass^[k,k] - FStiff^[k,k] * FMass^[j,j];
    C := (AB * AB + 4 * XX * YY) / 4.0;
    if C < 0.0 then
      begin
      if assigned (FOnNotPositive) then FOnNotPositive(Self);
      if D <> nil then begin Dispose (D); D:=nil; end;
      FActive:=False; exit;
      end;
    Q := Sqrt (C);
    D1 := AB/2.0 + Q;
    D2 := AB/2.0 - Q;
    Z := D1;
    if Abs(D2)>Abs(D1) then Z := D2;
    if Z = 0.0 then
      begin
      CA := 0.0;
      CG := - FStiff^[j,k] / FStiff^[k,k];
      end
    else
      begin
      CA := XX / Z;
      CG := - YY / Z;
      end;
    if FNumber <> 2 then
      begin
      j1 := j + 1;
      j2 := j - 1;
      k1 := k + 1;
      k2 := k - 1;
      if j2 >= 1 then
        for i := 1 to j2 do
          begin
          AJ := FStiff^[i,j];
          BJ := FMass^[i,j];
          AK := FStiff^[i,k];
          BK := FMass^[i,k];
          FStiff^[i,j] := AJ + CG * AK;
          FMass^[i,j] := BJ + CG * BK;
          FStiff^[i,k] := AK + CA * AJ;
          FMass^[i,k] := BK + CA * BJ;
          end;
      if k1 <= FNumber then
        for i := k1 to FNumber do
          begin
          AJ := FStiff^[j,i];
          BJ := FMass^[j,i];
          AK := FStiff^[k,i];
          BK := FMass^[k,i];
          FStiff^[j,i] := AJ + CG * AK;
          FMass^[j,i] := BJ + CG * BK;
          FStiff^[k,i] := AK + CA * AJ;
          FMass^[k,i] := BK + CA * BJ;
          end;
      if j1 <= k2 then
        for i := j1 to k2 do
          begin
          AJ := FStiff^[j,i];
          BJ := FMass^[j,i];
          AK := FStiff^[i,k];
          BK := FMass^[i,k];
          FStiff^[j,i] := AJ + CG * AK;
          FMass^[j,i] := BJ + CG * BK;
          FStiff^[i,k] := AK + CA * AJ;
          FMass^[i,k] := BK + CA * BJ;
          end;
      end;
    AK := FStiff^[k,k];
    BK := FMass^[k,k];
    FStiff^[k,k] := AK + 2.0 * CA * FStiff^[j,k] + CA * CA * FStiff^[j,j];
    FMass^[k,k] := BK + 2.0 * CA * FMass^[j,k] + CA * CA * FMass^[j,j];
    FStiff^[j,j] := FStiff^[j,j] + 2.0 * CG * FStiff^[j,k] + CG * CG * AK;
    FMass^[j,j] := FMass^[j,j] + 2.0 * CG * FMass^[j,k] + CG * CG * BK;
    FStiff^[j,k] := 0.0;
    FMass^[j,k] := 0.0;
    for i := 1 to FNumber do
      begin
      E := FForms^[i,j];
      F := FForms^[i,k];
      FForms^[i,j] := E + CG * F;
      FForms^[i,k] := F + CA * E;
      end;
Label2:
  end;
  end;
      for i := 1 to FNumber do
        begin
        if (FStiff^[i,i] <= 0.0) or (FMass^[i,i] <= 0.0) then
          begin
          if assigned (FOnNotPositive) then FOnNotPositive(Self);
          if D <> nil then
          begin
          Dispose (D);
          D:=nil;
          end;
          FActive:=False; Exit;
          end;
          FValues^[i] := FStiff^[i,i] / FMass^[i,i];
        end;
for i := 1 to FNumber do
begin
T := 0.0000001 * D^[i];
G := Abs (FValues^[i] - D^[i]);
if G > T then goto Label5;
end;
X := Sqr (0.0000001);
for j := 1 to M do
begin
jj := j + 1;
for k := jj to FNumber do
begin
XA := (FStiff^[j,k] * FStiff^[j,k]) / (FStiff^[j,j] * FStiff^[k,k]);
XB := (FMass^[j,k] * FMass^[j,k]) / (FMass^[j,j] * FMass^[k,k]);
if (XA < X) and (XB < X) then goto Label3;
goto Label5;
Label3:
end;
end;
Label4:
for i := 1 to FNumber do
for j := 1 to FNumber do
begin
FStiff^[j,i] := FStiff^[i,j];
FMass^[j,i] := FMass^[i,j];
end;
for j := 1 to FNumber do
begin
ZZ := Sqrt (FMass^[j, j]);
for k := 1 to FNumber do
FForms^[k,j] := FForms^[k,j] / ZZ;
end;
for j:=1 to FNumber do
begin
ZZ:=abs(FForms^[1,j]);
for i:=1 to FNumber do if(abs(FForms^[i,j])>ZZ) then ZZ:=abs(FForms^[i,j]);
for i:=1 to FNumber do FForms^[i,j]:=FForms^[i,j]/ZZ;
end;

for j:=1 to FNumber do
begin

tmp:=FValues^[j]; m:=j;

for k:=j to FNumber do
     if FValues^[k]<tmp then begin m:=k; tmp:=FValues^[k]; end;


tmp:=FValues^[m]; FValues^[m]:=FValues^[j]; FValues^[j]:=tmp;

for i:=1 to FNumber do
begin
tmp:=FForms^[i,m];
FForms^[i,m]:=FForms^[i,j];
FForms^[i,j]:=tmp;
end;

end;


if D <> nil then
begin Dispose (D); D:=nil; end;
Result:=True;
Exit;
Label5:
  for i := 1 to FNumber do
  D^[i] := FValues^[i];
  if L < 16 then goto Label1;
  goto Label4;
end;

end.

